﻿/*********************************************************
Copyright (C) 2018 Raymond Allen

MIT License:

┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
**********************************************************************/
/*********************************************************************

   Copyright (C) 2001 Smaller Animals Software, Inc.

   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.

   3. This notice may not be removed or altered from any source distribution.

   http://www.smalleranimals.com
   smallest@smalleranimals.com

   --------

   This code is based, in part, on:

      "Syntax Coloring Text Edit Window Class " - Randy More (9, May 1998)
      http://www.codeguru.com/editctrl/syntax_coloring.shtml

   Though probably only 5% of the original code remains.

**********************************************************************/

// Colorizer.cpp : implementation of the CColorizer class
//

#include "stdafx.h"
#include "Colorizer.h"
#include <string>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


/////////////////////////////////////////////////////////////////////////////

CColorizer::CColorizer()
{
	m_bCompareCase = TRUE;//RJA:We don't care about case of keywords in Spin... FALSE;
	m_sColorMode = "Spin2";//RJA:  Change to Spin1 mode with this: "Spin1"
}

/////////////////////////////////////////////////////////////////////////////

CColorizer::~CColorizer()
{
}

bool CColorizer::isalnumSpin(int c)
{//Need to treat "_" as part of alpha numeric
	if (c == '_')
		return true;
	return isalnum(c);

}

void CColorizer::SetLineColors(CStringA &csString, vector<COLORREF>& colors, COLORREF LineBackgroundColor, int nBraces) //rja removed const
{
   // nothing to do
   if (csString.GetLength() ==0) 
   {
      colors.clear();
      return;
   }

   // one COLORREF per input char
   colors.resize(csString.GetLength());


   //Dat sections are colored differently..
   bool bDatSection = false;
   if (LineBackgroundColor == RGB(191, 255, 200))
	   bDatSection = true;
   if (LineBackgroundColor == RGB(167, 253, 179))
	   bDatSection = true;
   //So are CON and VAR
   bool bVarConSection = false;
   if (LineBackgroundColor == RGB(255, 223, 191))
	   bVarConSection = true;
   if (LineBackgroundColor == RGB(253, 210, 167))
	   bVarConSection = true;
   if (LineBackgroundColor == RGB(253, 243, 168))
	   bVarConSection = true;
   if (LineBackgroundColor == RGB(255, 248, 192))
	   bVarConSection = true;
   //OBJ too
   bool bObjSection = false;
   if (LineBackgroundColor == RGB(255, 191, 191))
	   bObjSection = true;
   if (LineBackgroundColor == RGB(253, 167, 167))
	   bObjSection = true;

   // look for keywords

	CStringA word;
	word = csString;
	word.TrimLeft();
	word.TrimRight();
	word = word.Left(3);
	
	//set the line to all black
	int k = 0;
	int j = csString.GetLength();
	for(int i=0; i<j; i++)
	{//everything starts normal, except for '#' in DAT sections
		if (!bDatSection)
			colors.at(i) = NORMAL_COLOR;
		else
			if (csString[i] != '#')
				colors.at(i) = NORMAL_COLOR;
			else
				colors.at(i) = POUND_COLOR;

	}

   // do keyword coloring and auto-case correction
	while (k < j)
	{
		// find "words" (sequences of alpha-numeric characters)
		word.Empty();
		while ((k < j) && (!isalnumSpin((unsigned char)csString.GetAt(k)))) //RJA adding >127 check to avoid isalnum breaking
		{
			k++;
			while (csString.GetAt(k) > 127)
				k++;

		}

		int s = k;
		while ((k < j) && (isalnumSpin((unsigned char)csString.GetAt(k))))//RJA adding <128 check to avoid isalnum breaking
		{
			word = word + csString.GetAt(k);
			k++;
		}

		int e = k;

		word.TrimLeft();
		word.TrimRight();
		if (m_sColorMode != L"Spin2")
		{//spin1
			if (bDatSection == false)
			{//Not a DAT section 
				if (bVarConSection)
				{
					if (IsVarConKeyWord(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = KEYWORD_COLOR;
						}
					}
				}
				else if (bObjSection)
				{
					if (IsObjKeyWord(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = KEYWORD_COLOR;
						}
					}
				}
				else  //PUB or PRI
				{
					if (IsSpin1KeyWord(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = KEYWORD_COLOR;
						}
					}

					if (IsFlowControl(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = FLOW_COLOR;
						}
					}
				}




			}
			else
			{//DAT section
				if (word == "DAT")
				{//bold dat
					int l = 0;
					for (int i = s; i < e; i++)
					{
						l++;
						colors.at(i) = KEYWORD_COLOR;
					}
				}

				if (IsPasm1KeyWord(word))  //dat section keyword?
				{
					int l = 0;
					for (int i = s; i < e; i++)
					{
						l++;
						colors.at(i) = PASM_KEYWORD_COLOR;
					}
				}
				if (IsPasm1Reg(word))  //dat section register?
				{
					int l = 0;
					for (int i = s; i < e; i++)
					{
						l++;
						colors.at(i) = KEYWORD_COLOR;  //these are black like keywords in non-DAT sections
					}
				}

			}
		}
		else
		{//SPIN2
			if (bDatSection == false)
			{//Not a DAT section
				if (bVarConSection)
				{
					if (IsVarConKeyWord(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = KEYWORD_COLOR;
						}
					}
				}
				else if (bObjSection)
				{
					if (IsObjKeyWord(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = KEYWORD_COLOR;
						}
					}
				}
				else  //PUB or PRI
				{
					if (IsSpin1KeyWord(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = KEYWORD_COLOR;
						}
					}

					if (IsFlowControl(word))
					{
						int l = 0;
						for (int i = s; i < e; i++)
						{
							l++;
							colors.at(i) = FLOW_COLOR;
						}
					}
				}


			}
			else
			{//DAT section
				if (word == "DAT")
				{//bold dat
					int l = 0;
					for (int i = s; i < e; i++)
					{
						l++;
						colors.at(i) = KEYWORD_COLOR;
					}
				}

				if (IsPasm2KeyWord(word))  //dat section keyword?
				{
					int l = 0;
					for (int i = s; i < e; i++)
					{
						l++;
						colors.at(i) = PASM_KEYWORD_COLOR;
					}
				}
				if (IsPasm2Reg(word))  //dat section register?
				{
					int l = 0;
					for (int i = s; i < e; i++)
					{
						l++;
						colors.at(i) = KEYWORD_COLOR;  //these are black like keywords in non-DAT sections
					}
				}

			}
		}
	}

//RJA Do similar search for  for operators
		//set the line to all black
		k = 0;
		j = csString.GetLength();


		// do keyword coloring and auto-case correction
		while (k < j)
		{
			// find "words" (sequences of alpha-numeric characters)
			word.Empty();
			while ((k < j) && ((isblank((unsigned char)csString.GetAt(k)))||(isalnum((unsigned char)csString.GetAt(k))))) //RJA adding >127 check to avoid isalnum breaking
			{
				k++;
				while (csString.GetAt(k) > 127)
					k++;

			}

			int s = k;
			while ((k < j) && (!isblank((unsigned char)csString.GetAt(k))&&(!((isalnum((unsigned char)csString.GetAt(k)))))))//RJA adding <128 check to avoid isalnum breaking
			{
				word = word + csString.GetAt(k);
				k++;
			}

			int e = k;

			word.TrimLeft();
			word.TrimRight();

			if (IsOperator(word))
			{
				int l = 0;
				for (int i = s; i < e; i++)
				{
					l++;
					colors.at(i) = RGB(0, 128, 0);
				}
			}
		}

	
}

/////////////////////////////////////////////////////////////////////////////

BOOL CColorizer::IsSpin1KeyWord(CStringA & test)
{
   std::set<CStringA>::iterator it;
   if (m_bCompareCase)
   {
      CStringA t(test); t.MakeUpper();

      it = m_stringSet.find(t);
   }
   else
   {
      it = m_stringSet.find(test);
   }

   return (it != m_stringSet.end());
}

BOOL CColorizer::IsVarConKeyWord(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_stringSetVarCon.find(t);
	}
	else
	{
		it = m_stringSetVarCon.find(test);
	}

	return (it != m_stringSetVarCon.end());
}

BOOL CColorizer::IsObjKeyWord(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_stringSetObj.find(t);
	}
	else
	{
		it = m_stringSetObj.find(test);
	}

	return (it != m_stringSetObj.end());
}


BOOL CColorizer::IsPasm1KeyWord(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_stringSetPASM.find(t);
	}
	else
	{
		it = m_stringSetPASM.find(test);
	}

	return (it != m_stringSetPASM.end());
}

BOOL CColorizer::IsPasm1Reg(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_stringSetP1Regs.find(t);
	}
	else
	{
		it = m_stringSetP1Regs.find(test);
	}

	return (it != m_stringSetP1Regs.end());
}

BOOL CColorizer::IsSpin2KeyWord(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_stringSet2.find(t);
	}
	else
	{
		it = m_stringSet2.find(test);
	}

	return (it != m_stringSet2.end());
}

BOOL CColorizer::IsPasm2KeyWord(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_stringSetP2ASM.find(t);
	}
	else
	{
		it = m_stringSetP2ASM.find(test);
	}

	return (it != m_stringSetP2ASM.end());
}

BOOL CColorizer::IsPasm2Reg(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_stringSetP2Regs.find(t);
	}
	else
	{
		it = m_stringSetP2Regs.find(test);
	}

	return (it != m_stringSetP2Regs.end());
}

BOOL CColorizer::IsOperator(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_Operators.find(t);
	}
	else
	{
		it = m_Operators.find(test);
	}

	return (it != m_Operators.end());
}

BOOL CColorizer::IsFlowControl(CStringA & test)
{
	std::set<CStringA>::iterator it;
	if (m_bCompareCase)
	{
		CStringA t(test); t.MakeUpper();

		it = m_FlowControl.find(t);
	}
	else
	{
		it = m_FlowControl.find(test);
	}

	return (it != m_FlowControl.end());
}




/////////////////////////////////////////////////////////////////////////////

void CColorizer::LoadKeywordFile(CString &m_csaKeyWordsFile)
{
   if (m_csaKeyWordsFile.IsEmpty()) 
   {
      return;
   }

	FILE * infile;
	char buffer[512];
	CString inifile = m_csaKeyWordsFile;
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//RJA DAT section keywords
	CStringA skeyWords = "con var obj pub pri dat true false posx negx pi rcfast rcslow xinput xtal1 xtal2 xtal3 pll1x pll2x pll4x pll8x pll16x";
	skeyWords += " result quit return abort byte word long bytefill longfill wordfill bytemove wordmove longmove";
	skeyWords += " lookup lookupz lookdown lookdownz strsize strcomp string constant float round trunc file locknew lockret lockset lockclr";
	skeyWords += " chipver clkmode _clkmode clkfreq _clkfreq clkset _xinfreq _stack _free cnt ctra ctrb dira dirb ina inb outa outb frqa frqb phsa phsb vcfg vscl par spr";
	skeyWords += " waitcnt waitpeq waitpne waitvid cogid cognew coginit cogstop reboot";
	skeyWords.MakeUpper();
	
	CStringA resToken;
	int curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSet.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//RJA VAR and CON section keywords
	skeyWords = "con var obj pub pri dat true false posx negx pi rcfast rcslow xinput xtal1 xtal2 xtal3 pll1x pll2x pll4x pll8x pll16x";
	skeyWords += " float round trunc";
	skeyWords += " _clkmode _clkfreq _xinfreq _stack _free";
	skeyWords += " true false posx negx pi";

	skeyWords.MakeUpper();

	resToken;
	curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSetVarCon.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//RJA OBJ section keywords
	skeyWords = "float round trunc";
	
	skeyWords.MakeUpper();

	resToken;
	curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSetObj.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};


	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//RJA PASM Keywords:  m_stringSetPASM
	skeyWords = "fit res"; //ORG is actually colored black...  "org fit res ";
	skeyWords += " clkset";
	skeyWords += " cogid coginit cogstop";
	skeyWords += " locknew lockret lockclr lockset waitcnt waitpeq waitpne waitvid";
	skeyWords += " if_always if_never if_e if_ne if_a if_b if_ae if_be if_c if_nc if_z if_nz";
	skeyWords += " if_c_eq_z if_c_ne_z if_c_and_z if_c_and_nz if_nc_and_z if_nc_and_z if_nc_and_nz";
	skeyWords += " if_z_eq_c if_z_ne_c if_z_and_c if_z_and_nc if_nz_and_c if_nz_and_nc";
	skeyWords += " if_z_or_c if_z_or_nc if_nz_or_c if_nz_or_nc";
	skeyWords += " call djnz jmp jmpret tjnz tjz ret";
	skeyWords += " nr wr wc wz";
	skeyWords += " rdbyte rdword rdlong wrbyte wrword wrlong";
	skeyWords += " abs absneg neg negc negnc negz negnz min mins max maxs";
	skeyWords += " add addabs adds addx addsx";
	skeyWords += " sub subabs subs subx subsx";
	skeyWords += " sumc sumnc sumz sumnz mul muls";
	skeyWords += " and andn or xor ones enc rcl rcr rev rol ror shl shr sar";
	skeyWords += " cmp cmps cmpx cmpsx cmpsub test testn mov movs movd movi";
	skeyWords += " muxc muxnc muxz muxnz hubop nop";
	skeyWords += " true false posx negx pi";
	skeyWords.MakeUpper();

	curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSetPASM.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//Spin2 keywords  (needs input from Chip!)
	skeyWords = "con var obj pub pri dat true false posx negx pi rcfast rcslow xinput xtal1 xtal2 xtal3 pll1x pll2x pll4x pll8x pll16x";
	skeyWords += " result quit return abort byte word long bytefill longfill wordfill bytemove wordmove longmove";
	skeyWords += " lookup lookupz lookdown lookdownz strsize strcomp string constant float round trunc file locknew lockret lockset lockclr";
	skeyWords += " chipver clkmode _clkmode clkfreq _clkfreq clkset _xinfreq _stack _free cnt ctra ctrb dira dirb ina inb outa outb spr";
	skeyWords += " waitcnt waitpeq waitpne waitvid cogid cognew coginit cogstop reboot";
	skeyWords.MakeUpper();

	resToken;
	curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSet2.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//RJA PASM Registers:  m_stringSetP1Regs
	skeyWords = "org file res fit long byte word"; //not really registers, but still black
	skeyWords += " dira dirb ina inb outa outb";
	skeyWords += " cnt ctra ctrb frqa frqb phsa phsb vcfg vscl par";
	skeyWords.MakeUpper();

	curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSetP1Regs.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};

	//P2ASM Keywords
	skeyWords = "NOP ROR ROL SHR SHL RCR RCL SAR SAL ADD ADDX ADDS ADDSX SUB SUBX SUBS SUBSX CMP CMPX CMPS CMPSX CMPR CMPM SUBR CMPSUB FGE FLE FGES FLES SUMC SUMNC SUMZ SUMNZ TESTB TESTBN BITL BITH BITC BITNC BITZ BITNZ BITRND BITNOT AND ANDN OR XOR MUXC MUXNC MUXZ";
	skeyWords += " MUXNZ MOV NOT NOT ABS ABS NEG NEG NEGC NEGC NEGNC NEGNC NEGZ NEGZ NEGNZ NEGNZ INCMOD DECMOD ZEROX SIGNX ENCOD ENCOD ONES ONES TEST TEST TESTN SETNIB SETNIB GETNIB GETNIB ROLNIB ROLNIB SETBYTE SETBYTE GETBYTE GETBYTE ROLBYTE ROLBYTE SETWORD GETWORDROLWORD ALTSN ALTGN";
	skeyWords += " ALTSB ALTGB ALTSW ALTGW ALTR ALTD ALTS ALTB ALTI SETR SETD SETS DECOD BMASK CRCBIT CRCNIB MUXNITS MUXNIBS MUXQ MOVBYTS MUL MULS SCA SCAS ADDPIX MULPIX BLNPIX MIXPIX ADDCT1 ADDCT2 ADDCT3 WMLONG RQPIN RDPIN RDLUT RDBYTE RDWORD RDLONG POPA POPB";
	skeyWords += " CALLD RESI3 RESI2 RESI1 RESI0 CALLPA CALLPB DJZ DJNZ DJF DJNF IJZ IJNZ TJZ TJNZ TJF TJNF TJS TJNS TJV JINT JCT1 JCT2 JCT3 JSE1 JSE2 JSE3 JSE4 JPAT JFBW JXMT JXFI JXRO JXRL JATN JQMT JNINT JNCT1 JNCT2 JNCT3 JNSE1 JNSE2 JNSE3 JNSE4 JNPAT";
	skeyWords += " JNFBW JNXMT JNXFI JNXRO JNXRL JNATN JNQMT SETPAT WRPIN AKPIN WXPIN WYPIN WRLUT WRBYTE WRWORD WRLONG PUSHA PUSHB RDFAST WRFAST FBLOCK XINIT XSTOPXZERO XCONT REP COGINIT QMUL QDIV QFRAC QSQRT QROTATE QVECTOR HUBSET COGID COGSTOP LOCKNEW LOCKRET LOCKTRY LOCKREL QLOG QEXP RFBYTE RFWORD RFLONG RFVAR RFVARS WFBYTE WFWORD WFLONG GETQX GETQY";
	skeyWords += " GETCT GETRND SETDACS SETXFRQ GETXACC WAITX SETSE1 SETSE2 SETSE3 SETSE4 POLLINT POLLCT1 POLLCT2 POLLCT3 POLLSE1 POLLSE2 POLLSE3 POLLSE4 POLLPAT POLLFBW POLLXMT POLLXFI POLLXRO POLLXRL POLLATN POLLQMT WAITINT WAITCT1 WAITCT2 WAITCT3 WAITSE1 WAITSE2 WAITSE3 WAITSE4 WAITPAT WAITFBW WAITXMT WAITXFI WAITXRO WAITXRL WAITATN ALLOWI STALLI TRGINT1 TRGINT2 TRGINT3 NIXINT1 NIXINT2 NIXINT3 SETINT1";
	skeyWords += " SETINT2 SETINT3 SETQ SETQ2 PUSH POP JMP CALL RET CALLA RETA CALLB RETB JMPREL SKIP SKIPF EXECF GETPTR GETBRK COGBRK BRK SETLUTS LUTSOFF LUTSON SETCY SETCI SETCQ SETCFRQ SETCMOD SETPIV SETPIX COGATN TESTP TESTPN DIRL DIRH DIRC DIRNC DIRZ DIRNZ DIRRND DIRNOT OUTL OUTH OUTC OUTNC OUTZ OUTNZ OUTRND OUTNOT";
	skeyWords += " FLTL FLTH FLTC FLTNC FLTZ FLTNZ FLTRND FLTNOT DRVL DRVH DRVC DRVNC DRVZ DRVNZ DRVRND DRVNOT SPLITB MERGEB SPLITW MERGEW SEUSSF SEUSSR RGBSQZ RGBEXP XORO32 REV RCZR RCZL WRC WRNC WRZ WRNZ MODCZ MODC MODZ JMP CALL CALLA CALLB CALLD LOC AUGS AUGD";
	skeyWords += " _RET_ IF_NC_AND_NZ IF_NZ_AND_NC IF_A IF_00 IF_NC_AND_Z IF_Z_AND_NC IF_01 IF_NC IF_AE IF_0X IF_C_AND_NZ IF_NZ_AND_C IF_10 IF_NZ IF_NE IF_X0 IF_C_NE_Z IF_Z_NE_C IF_DIFF IF_NC_OR_NZ IF_NZ_OR_NC IF_NOT_11 IF_C_AND_Z IF_Z_AND_C IF_11 IF_C_EQ_Z IF_Z_EQ_C IF_SAME IF_Z IF_E IF_X1 IF_NC_OR_Z IF_Z_OR_NC IF_NOT_10 IF_C IF_B IF_1X IF_C_OR_NZ IF_NZ_OR_C IF_NOT_01 IF_C_OR_Z IF_Z_OR_C IF_BE IF_NOT_00";
	skeyWords.MakeUpper();

	curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSetP2ASM.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};
	//////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//RJA PASM Registers:  m_stringSetP1Regs
	skeyWords = "org orgh orgf file res fit alignw alignl long byte word"; //not really registers, but still black
	skeyWords += " _CLR _NC_AND_NZ _NZ_AND_NC _GT _NC_AND_Z _Z_AND_NC _NC _GE _C_AND_NZ _NZ_AND_C _NZ _NE _C_NE_Z _Z_NE_C _NC_OR_NZ _NZ_OR_NC _C_AND_Z _Z_AND_C _C_EQ_Z _Z_EQ_C _Z _E _NC_OR_Z _Z_OR_NC _C _LT _C_OR_NZ _NZ_OR_C _C_OR_Z _Z_OR_C _LE _SET";
	skeyWords += " dira dirb ina inb outa outb ptra ptrb PA PB IJMP1 IJMP2 IJMP3 IRET1 IRET2 IRET3";
	//skeyWords += "par";
	skeyWords.MakeUpper();

	curPos = 0;

	resToken = skeyWords.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_stringSetP2Regs.insert(resToken);
		resToken = skeyWords.Tokenize(" ", curPos);
	};

	/*RJA non-unicode
	if(infile = fopen(inifile,"rb"))
	{
		while(!feof(infile))
		{
			fscanf(infile,"%s",buffer);

         CStringA t(buffer);
         if (m_bCompareCase)
         {
            t.MakeUpper();
         }

         m_stringSet.insert(t);
		}

		fclose(infile);
	}
	*/
}

void CColorizer::LoadFlowControl()
{
	CStringA sFlow = "if else elseif elseifnot ifnot case other repeat from to step until while next";
	sFlow.MakeUpper();
	CStringA resToken;
	int curPos = 0;

	resToken = sFlow.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_FlowControl.insert(resToken);
		resToken = sFlow.Tokenize(" ", curPos);
	};
}


void CColorizer::LoadOperators()
{
	CStringA sOperators = "-- ++ ? @ @@ ^^ || |< >| ! : ~~ ~";
	sOperators += "<-  <-= ->  ->= <<  <<= >>  >>= ->  ->= ><  ><= &   &= |   |= ^   ^= *   *= **  **= /   /= //  //= +   += -   -= #>  #>= <#  <#= <   <= >   >= <>  <>=";
	sOperators += " ==  === =<  =<= =>  =>= =   := not and and= or or= NOT AND AND= OR OR=";

	CStringA resToken;
	int curPos = 0;

	resToken = sOperators.Tokenize(" ", curPos);
	while (resToken != _T(""))
	{
		m_Operators.insert(resToken);
		resToken = sOperators.Tokenize(" ", curPos);
	};
}

