﻿/*********************************************************
Copyright (C) 2018 Raymond Allen

MIT License:

┌──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┐
│                                                   TERMS OF USE: MIT License                                                  │
├──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┤
│Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    │
│files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    │
│modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software│
│is furnished to do so, subject to the following conditions:                                                                   │
│                                                                                                                              │
│The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.│
│                                                                                                                              │
│THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          │
│WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         │
│COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   │
│ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         │
└──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────┘
**********************************************************************/
/*********************************************************************

   Copyright (C) 2001 Smaller Animals Software, Inc.

   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
   arising from the use of this software.

   Permission is granted to anyone to use this software for any purpose,
   including commercial applications, and to alter it and redistribute it
   freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.

   3. This notice may not be removed or altered from any source distribution.

   http://www.smalleranimals.com
   smallest@smalleranimals.com

   --------

   This code is based, in part, on:

      "Syntax Coloring Text Edit Window Class " - Randy More (9, May 1998)
      http://www.codeguru.com/editctrl/syntax_coloring.shtml

   Though probably only 5% of the original code remains.

**********************************************************************/

#include "stdafx.h"
#include "ScriptText.h"
/////////////////////////////////////////////////////////////////////////////


CScriptText::CScriptText ()
{
   m_pColorizer = &m_defaultColorizer;

    m_maxLineLen = 0;

}

/////////////////////////////////////////////////////////////////////////////

CScriptText & CScriptText::operator = (CScriptText & source)
{
   m_textArray.RemoveAll();
   m_textArray.Copy(source.m_textArray);
   m_colors = source.m_colors;

   return *this;
}

/////////////////////////////////////////////////////////////////////////////

void CScriptText::_InsertLine(CStringA &line, int i, BOOL bUpdate)  //rja removed const
{
   m_textArray.InsertAt(i, line); 
   lineVector lv;

   // sometimes, we don't want to update each line. like on a large paste.
   // we'll update all, later.
   if (bUpdate)
   {
      FindMaxLineLen();
	  if (SetLineBackgroundColors())
		  bUpdate=true;//Need to increase size of m_LineBackgroundColors
	  int nBraces = (i > 0) ? m_LineBracesCount[i - 1] : 0;  //get brace count from last line
      m_pColorizer->SetLineColors(line, lv, m_LineBackgroundColors[i],nBraces);
   }
   vector<CScriptText::lineVector>::iterator j;
   j = m_colors.begin();
   j += i;
   m_colors.insert((vector<CScriptText::lineVector>::iterator) j, lv);
}

/////////////////////////////////////////////////////////////////////////////

void CScriptText::_SetLine(CStringA &line, int i)   //rja removed const  
{
   m_textArray.SetAt(i, line); 
   int nBraces = (i > 0) ? m_LineBracesCount[i - 1] : 0;  //get brace count from last line
   m_pColorizer->SetLineColors(line, m_colors.at(i), m_LineBackgroundColors[i], nBraces);
   FindMaxLineLen();
}

/////////////////////////////////////////////////////////////////////////////

void CScriptText::_RemoveLine(int i)                       
{
   m_textArray.RemoveAt(i);       

   m_colors.erase(m_colors.begin() + i);

   FindMaxLineLen();
}

/////////////////////////////////////////////////////////////////////////////

void CScriptText::_RemoveAll()                             
{
   m_textArray.RemoveAll(); 
   m_colors.clear();
   FindMaxLineLen(); 

   CStringA firstLine;
   _InsertLine(firstLine, 0, TRUE);
}

/////////////////////////////////////////////////////////////////////////////

long CScriptText::_GetSize() const                               
{
   return m_textArray.GetSize();
}

/////////////////////////////////////////////////////////////////////////////

CStringA CScriptText::_GetLine(int i) const                       
{
   return m_textArray[i];
}

/////////////////////////////////////////////////////////////////////////////

const CScriptText::lineVector &CScriptText::_GetLineColors(int i) const                       
{
   return m_colors.at(i);
}

/////////////////////////////////////////////////////////////////////////////

void  CScriptText::FullUpdate()
{
   FindMaxLineLen();
   SetLineBackgroundColors();//rja
   for (int i=0;i<_GetSize();i++)
   {
	  int nBraces = (i > 0) ? m_LineBracesCount[i - 1] : 0;  //get brace count from last line
      m_pColorizer->SetLineColors(_GetLine(i), m_colors.at(i), m_LineBackgroundColors[i],nBraces);
   }
}

/////////////////////////////////////////////////////////////////////////////

long  CScriptText::FindMaxLineLen()
{
   m_maxLineLen = 0;
   for (int i=0;i<m_textArray.GetSize();i++)
   {
      m_maxLineLen = max(m_maxLineLen, m_textArray[i].GetLength());
   }
   
   return m_maxLineLen;
}

/////////////////////////////////////////////////////////////////////////////

void CScriptText::SetColorizer(CColorizer *pColorizer)
{
   m_pColorizer = pColorizer;
}

void CScriptText::InitLineBackgroundColors()
{//RJA since view is setting initial text, need to give it 
}

bool CScriptText::SetLineBackgroundColors()  //RJA:  NOTE:  ToDo:  Support multiple background colors on a single line.  Perhaps make array indicating whether each line is a special case or not...
{//RJA:  Pick background colors for every line, return value indicates whether should invalide or not
	//Also:  Calculate a braces count for start of each line

	//count sections


	//See if new lines were added since last time
	bool bRedraw = false;
	int n = _GetSize();// GetTextSize();
	if (m_LineBackgroundColors.GetSize() != n)  //if number of lines has changed, need to redraw window and resize
	{
		m_LineBackgroundColors.SetSize(n);
		bRedraw = true;
	}
	if (m_LineBracesCount.GetSize() != n)  //Do the same for braces count
	{
		m_LineBracesCount.SetSize(n);
		bRedraw = true;
		//RJA:  For now, assume these other arrays only need resizing here too
		m_nCon.SetSize(n);
		m_nVar.SetSize(n);
		m_nObj.SetSize(n);
		m_nPub.SetSize(n);
		m_nPri.SetSize(n);
		m_nDat.SetSize(n);
	}

	//Init running section counts
	//m_nCon[0] = 0; m_nVar[0] = 0; m_nObj[0] = 0; m_nPub[0] = 0; m_nPri[0] = 0; m_nDat[0] = 0;

	//Calculate running brace count
	m_LineBracesCount[0] = 0;
	int nBraces = 0;
	for (int i = 0; i < n; i++)
	{
		CStringA line = _GetLine(i); //GetTextLine(i);
		line.TrimLeft();
		int l = line.Find("'", 0); //need to ignore everything after comment
		if (l < 0)
			l = n;
		//for (int j = 0; j < l; j++)
		//{
			int b = line.Find("{", 0);
			while ((b >= 0)&&(b<l))
			{
				nBraces++;
				b = line.Find("{", b+1);

			}
			b = line.Find("}", 0);
			while ((b >= 0) && (b < l))
			{
				nBraces--;
				b = line.Find("}", b+1);

			}
			if (m_LineBracesCount[i] != nBraces)
			{
				bRedraw = true;
				m_LineBracesCount[i] = nBraces;  //for drawing, need to look at past line
			}

		//}
	}


	COLORREF currentcolor = RGB(253, 243, 168);
	COLORREF newcolor;
	for (int i = 0; i < n; i++)
	{
		//if this line is braced out, can just continue
		int nLastBraces = (i>0) ? m_LineBracesCount[i - 1] : 0;
		nBraces = m_LineBracesCount[i];
		if ((i > 0) && (nLastBraces > 0)&&(nBraces>0))
		{//This line is braced out and not unbraced
			if (m_LineBackgroundColors[i] != currentcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = currentcolor;
			continue;//no need to look further
		}

		CStringA line = _GetLine(i); //GetTextLine(i);

		//If braces end on this line, need to strip to brace
		if ((nLastBraces > 0) && (nBraces == 0))
		{
			int b = line.Find("}", 0);
			while (b >= 0)
			{
				b++;
				nLastBraces--;
				if (nLastBraces == 0)
				{
					line = line.Right(line.GetLength() - b);
					break;
				}
				b = line.Find("}", b);
					
			}

		}

		line.TrimLeft();
		int l = line.FindOneOf(" '");//RJA Needs to include apostrophe in this search
		if (l > 0)
		{//get everything to left of first space
			line = line.Mid(0, l);
		}
		else
		{//no spaces on this line
			if (m_LineBackgroundColors[i] != currentcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = currentcolor;
		}

		line.MakeUpper();

		//init section count to last lines value
		m_nPub[i] = (i == 0) ? 0 : m_nPub[i - 1];
		m_nPri[i] = (i == 0) ? 0 : m_nPri[i - 1];
		m_nCon[i] = (i == 0) ? 0 : m_nCon[i - 1];
		m_nVar[i] = (i == 0) ? 0 : m_nVar[i - 1];
		m_nDat[i] = (i == 0) ? 0 : m_nDat[i - 1];
		m_nObj[i] = (i == 0) ? 0 : m_nObj[i - 1];

		if (line == "PUB")
		{
			m_nPub[i]++;
			newcolor = (m_nPub[i] & 1) ? RGB(191, 223, 255) : RGB(167, 210, 253);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "PRI")
		{
			m_nPri[i]++;
			newcolor = (m_nPri[i] & 1) ? RGB(191, 248, 255) : RGB(167, 243, 253);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "CON")
		{
			m_nCon[i]++;
			newcolor = (m_nCon[i] & 1) ? RGB(253, 243, 168) : RGB(255, 248, 192);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "VAR")
		{
			m_nVar[i]++;
			newcolor = (m_nVar[i] & 1) ? RGB(255, 223, 191) : RGB(253, 210, 167);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "DAT")
		{
			m_nDat[i]++;
			newcolor = (m_nDat[i] & 1) ? RGB(191, 255, 200) : RGB(167, 253, 179);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "OBJ")
		{
			m_nObj[i]++;
			newcolor = (m_nObj[i] & 1) ? RGB(255, 191, 191) : RGB(253, 167, 167);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else
		{
			if (m_LineBackgroundColors[i] != currentcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = currentcolor;
		}
		currentcolor = m_LineBackgroundColors[i];
	}
	return bRedraw;
	//if (bRedraw == true)
	//	Invalidate();

}

bool CScriptText::SetLineBackgroundColors(int nLine) //RJA:  Trying to make inserting text faster by only updating background colors of nearby lines
//RJA:  NOTE:  ToDo:  Support multiple background colors on a single line.  Perhaps make array indicating whether each line is a special case or not...
{//RJA:  Pick background colors for nearby lines, return value indicates whether should invalidate or not
	//Also:  Calculate a braces count for start of each line

	//count sections


	//See if new lines were added since last time
	bool bRedraw = false;
	int n = _GetSize();// GetTextSize();
	if (m_LineBackgroundColors.GetSize() != n)  //if number of lines has changed, need to redraw window and resize
	{
		m_LineBackgroundColors.SetSize(n);
		bRedraw = true;
	}
	if (m_LineBracesCount.GetSize() != n)  //Do the same for braces count
	{
		m_LineBracesCount.SetSize(n);
		bRedraw = true;
		//RJA:  For now, assume these other arrays only need resizing here too
		m_nCon.SetSize(n);
		m_nVar.SetSize(n);
		m_nObj.SetSize(n);
		m_nPub.SetSize(n);
		m_nPri.SetSize(n);
		m_nDat.SetSize(n);
	}

	int nStartLine = nLine - 1;  //start one line back
	int nEndLine = nLine + 1;
	if (nStartLine < 1)
		nStartLine = 0;
	
	

	//Calculate running brace count
	int nBraces = m_LineBracesCount[nStartLine];
	for (int i = nStartLine; i < nEndLine; i++)
	{
		CStringA line = _GetLine(i); //GetTextLine(i);
		line.TrimLeft();
		int l = line.Find("'", 0); //need to ignore everything after comment
		if (l < 0)
			l = n;
		//for (int j = 0; j < l; j++)
		//{
		int b = line.Find("{", 0);
		while ((b >= 0) && (b < l))
		{
			nBraces++;
			b = line.Find("{", b + 1);

		}
		b = line.Find("}", 0);
		while ((b >= 0) && (b < l))
		{
			nBraces--;
			b = line.Find("}", b + 1);

		}
		if (m_LineBracesCount[i] != nBraces)
		{
			bRedraw = true;
			m_LineBracesCount[i] = nBraces;  //for drawing, need to look at past line
		}

		//}
	}


	COLORREF currentcolor = m_LineBackgroundColors[nStartLine];// RGB(253, 243, 168);
	COLORREF newcolor;
	for (int i = nStartLine; i < nEndLine; i++)
	{
		//if this line is braced out, can just continue
		int nLastBraces = (i > 0) ? m_LineBracesCount[i - 1] : 0;
		nBraces = m_LineBracesCount[i];
		if ((i > 0) && (nLastBraces > 0) && (nBraces > 0))
		{//This line is braced out and not unbraced
			if (m_LineBackgroundColors[i] != currentcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = currentcolor;
			continue;//no need to look further
		}

		CStringA line = _GetLine(i); //GetTextLine(i);

		//If braces end on this line, need to strip to brace
		if ((nLastBraces > 0) && (nBraces == 0))
		{
			int b = line.Find("}", 0);
			while (b >= 0)
			{
				b++;
				nLastBraces--;
				if (nLastBraces == 0)
				{
					line = line.Right(line.GetLength() - b);
					break;
				}
				b = line.Find("}", b);

			}

		}

		line.TrimLeft();
		int l = line.FindOneOf(" '");//RJA Needs to include apostrophe in this search
		if (l > 0)
		{//get everything to left of first space
			line = line.Mid(0, l);
		}
		else
		{//no spaces on this line
			if (m_LineBackgroundColors[i] != currentcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = currentcolor;
		}

		line.MakeUpper();

		//init section count to last lines value
		m_nPub[i] = (i == 0) ? 0 : m_nPub[i - 1];
		m_nPri[i] = (i == 0) ? 0 : m_nPri[i - 1];
		m_nCon[i] = (i == 0) ? 0 : m_nCon[i - 1];
		m_nVar[i] = (i == 0) ? 0 : m_nVar[i - 1];
		m_nDat[i] = (i == 0) ? 0 : m_nDat[i - 1];
		m_nObj[i] = (i == 0) ? 0 : m_nObj[i - 1];

		if (line == "PUB")
		{
			m_nPub[i]++;
			newcolor = (m_nPub[i] & 1) ? RGB(191, 223, 255) : RGB(167, 210, 253);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "PRI")
		{
			m_nPri[i]++;
			newcolor = (m_nPri[i] & 1) ? RGB(191, 248, 255) : RGB(167, 243, 253);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "CON")
		{
			m_nCon[i]++;
			newcolor = (m_nCon[i] & 1) ? RGB(253, 243, 168) : RGB(255, 248, 192);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "VAR")
		{
			m_nVar[i]++;
			newcolor = (m_nVar[i] & 1) ? RGB(255, 223, 191) : RGB(253, 210, 167);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "DAT")
		{
			m_nDat[i]++;
			newcolor = (m_nDat[i] & 1) ? RGB(191, 255, 200) : RGB(167, 253, 179);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else if (line == "OBJ")
		{
			m_nObj[i]++;
			newcolor = (m_nObj[i] & 1) ? RGB(255, 191, 191) : RGB(253, 167, 167);
			if (m_LineBackgroundColors[i] != newcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = newcolor;
		}
		else
		{
			if (m_LineBackgroundColors[i] != currentcolor)
				bRedraw = true;
			m_LineBackgroundColors[i] = currentcolor;
		}
		currentcolor = m_LineBackgroundColors[i];
	}
	return bRedraw;
	//if (bRedraw == true)
	//	Invalidate();

}
